"""Test bluepymm/prepare_combos"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os

from bluepymm import tools, prepare_combos


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/prepare_combos')


def _verify_emodel_json(filename, output_dir, nb_emodels):
    """Helper function to verify emodel json output"""
    data_json = os.path.join(output_dir, filename)
    assert os.path.isfile(data_json)
    data = tools.load_json(data_json)
    assert len(data) == nb_emodels
    return data


def _verify_prepare_combos_output(scores_db, emodels_hoc_dir, output_dir,
                                  nb_emodels):
    """Helper function to verify output of prepare combos"""
    # TODO: test database contents
    assert os.path.isfile(scores_db)

    assert os.path.isdir(emodels_hoc_dir)
    hoc_files = os.listdir(emodels_hoc_dir)
    assert len(hoc_files) == nb_emodels
    for hoc_file in hoc_files:
        assert hoc_file.endswith('.hoc')

    _verify_emodel_json('final.json', output_dir, nb_emodels)
    emodel_dirs = _verify_emodel_json('emodel_dirs.json', output_dir,
                                      nb_emodels)
    for emodel in emodel_dirs:
        assert os.path.isdir(emodel_dirs[emodel])


def _prepare_config_json(original_filename, test_dir):
    """Helper function to prepare new configuration file."""
    config = tools.load_json(original_filename)
    config['tmp_dir'] = os.path.join(test_dir, 'tmp')
    config['output_dir'] = os.path.join(test_dir, 'output')
    config['scores_db'] = os.path.join(config['output_dir'], 'scores.sqlite')
    config['emodels_hoc_dir'] = os.path.join(config['output_dir'],
                                             'emodels_hoc')
    tools.makedirs(test_dir)
    return tools.write_json(test_dir, 'config.json', config)


def _test_prepare_combos(test_data_dir, config_template_path, nb_emodels,
                         test_dir):
    """Helper function to perform functional test prepare_combos"""
    with tools.cd(test_data_dir):
        # prepare new configuration file based on test_dir
        config_path = _prepare_config_json(config_template_path, test_dir)

        # run combination preparation
        prepare_combos.main.prepare_combos(conf_filename=config_path,
                                           continu=False)

        # test output
        config = tools.load_json(config_path)
        _verify_prepare_combos_output(config['scores_db'],
                                      config['emodels_hoc_dir'],
                                      config['output_dir'], nb_emodels)


def test_prepare_combos_from_dir():
    """bluepymm.prepare_combos: test prepare_combos with plain directory input
    based on example simple1
    """
    config_template_path = 'simple1_conf_prepare.json'
    nb_emodels = 2
    test_dir = os.path.join(TMP_DIR, 'test_prepare_combos_from_dir')

    _test_prepare_combos(TEST_DATA_DIR, config_template_path, nb_emodels,
                         test_dir)


def test_prepare_combos_from_git_repo():
    """bluepymm.prepare_combos: test prepare_combos with git repo input
    based on example simple1
    """
    config_template_path = 'simple1_conf_prepare_git.json'
    nb_emodels = 2
    test_dir = os.path.join(TMP_DIR, 'test_prepare_combos_from_git_repo')

    _test_prepare_combos(TEST_DATA_DIR, config_template_path, nb_emodels,
                         test_dir)



"""Functions for BluePyMM reporting."""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


# pylint: disable=R0914, C0325, W0640, W0633
# pylama: ignore=E402

import os

import pandas
import numpy

import matplotlib
matplotlib.use('Agg')
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
plt.style.use('ggplot')

from . import table_processing
from bluepymm import tools


BLUE = 'C1'
RED = 'C0'
YELLOW = 'C4'
FIGSIZE = (15, 10)


def pdf_file(pdf_filename):
    """Create and return a PDF file.

    Args:
        pdf_filename: path to PDF file

    Returns:
        A multi-page PDF file.
    """
    tools.makedirs(os.path.dirname(pdf_filename))
    return PdfPages(pdf_filename)


def add_plot_to_report(pp, plot_function, *args):
    """Add a plot to a given report.

    Args:
        pp: pdf file
        plot_function: function that returns figure
        args: arguments to plot_function
    """
    fig = plot_function(*args)
    pp.savefig(fig, bbox_inches='tight')
    plt.close()


def plot_dict(dict_data, title):
    """Plot a dictionary.

    Args:
        dict_data: a dictionary
        title: string with plot title

    Returns:
        Figure with plotted dictionary
    """
    fig = plt.figure(figsize=FIGSIZE)
    plt.axis('off')
    if dict_data:
        plt.table(
            cellText=[[x] for x in dict_data],
            loc='center')
    plt.title(title)
    plt.tight_layout()
    return fig


def plot_stacked_bars(
        data,
        xlabel,
        ylabel,
        title,
        color_map,
        log=False,
        yticksize=None):
    """Plot stacked bars.

    Args:
        data: a pandas.DataFrame
        xlabel: string with label for x-axis
        ylabel: string with label for y-axis
        title: string with plot title
        color_map: list of colors

    Returns:
        Figure with plot of stacked bars
    """
    ax = data.plot(
        kind='barh',
        figsize=FIGSIZE,
        stacked=True,
        color=color_map,
        log=log)
    if not log:
        ax.get_xaxis().set_major_locator(
            matplotlib.ticker.MaxNLocator(integer=True))
    else:
        plt.xlim(xmin=0.1)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if yticksize is not None:
        plt.yticks(fontsize=yticksize)
    plt.title(title)
    plt.tight_layout()
    plt.legend(loc='upper right')
    return ax.get_figure()


def plot_morphs_per_feature_for_emodel(emodel, megate_scores,
                                       emodel_score_values):
    """Display number of tested morphologies per feature for a given e-model.

    Args:
        emodel: string representing e-model, used for plot title
        megate_scores: pandas.DataFrame with megate scores, one entry per run
                       combo
        emodel_score_values: pandas.DataFrame with score values, one entry per
                             run combo

    Returns:
        Figure with plot of stacked bars. Passed and failed simulations are
        colored blue and red, respectively.
    """
    sums = pandas.DataFrame()
    sums['passed'] = megate_scores.sum(axis=0)
    sums['failed'] = len(emodel_score_values) - sums['passed']

    return plot_stacked_bars(
        sums, '# morphologies', '',
        '{}: number of tested morphologies per feature'.format(emodel),
        [BLUE, RED])


def plot_morphs_per_mtype_for_emodel(emodel, fullmtypes, megate_scores):
    """Display number of tested morphologies per m-type for a given e-model.

    Args:
        emodel: string representing e-model, used for plot title
        fullmtypes: pandas.DataFrame with m-types, one entry per run combo
        megate_scores: pandas.DataFrame with megate scores, one entry per run
                       combo

    Returns:
        Figure with plot of stacked bars. Passed and failed simulations are
        colored blue and red, respectively.
    """
    sums = pandas.DataFrame()
    for mtype in fullmtypes.unique():
        megate_scores_mtype = megate_scores[fullmtypes == mtype]
        mtype_passed = megate_scores_mtype[megate_scores_mtype['Passed all']]
        sums.loc[mtype, 'passed'] = len(mtype_passed)
        sums.loc[mtype, 'failed'] = (len(megate_scores_mtype) -
                                     sums.loc[mtype, 'passed'])

    return plot_stacked_bars(
        sums, '# morphologies', '',
        '{}: number of tested morphologies per m-type'.format(emodel),
        [BLUE, RED])


def create_morphology_label(data_frame):
    """Create label for morphology.

    Args:
        data_frame: pandas.DataFrame with columns 'morph_name', 'fullmtype',
                    and 'etype'

    Returns:
        A label (string), based on the contents of the first row of
        `data_frame`: <morph_name> (<fullmtype>, <etype>).
    """
    morph = data_frame.iloc[0]['morph_name']
    mtype = data_frame.iloc[0]['fullmtype']
    etype = data_frame.iloc[0]['etype']
    return '{} ({}, {})'.format(morph, mtype, etype)


def plot_emodels_per_morphology(data, final_db):
    """Display result of tested e-models for each morphology.

    Args:
        data: pandas.DataFrame with data on run combos
        final_db: pandas.DataFrame with data on selected combos

    Returns:
        Figure with plot of stacked bars. Simulations that passed, threw an
        error, and failed are colored blue, yellow and red, respectively.
    """
    sums = pandas.DataFrame()
    non_exemplars = data[data['is_exemplar'] == 0]
    for morph_name in non_exemplars['morph_name'].unique():
        nb_matches = len(final_db[final_db['morph_name'] == morph_name])
        non_exemplars_morph = non_exemplars[
            non_exemplars['morph_name'] == morph_name]
        nb_errors = len(
            non_exemplars_morph[non_exemplars_morph['exception'].notnull()])
        nb_combos = len(non_exemplars_morph)

        label = create_morphology_label(non_exemplars_morph)
        sums.loc[label, 'passed'] = nb_matches
        sums.loc[label, 'error'] = nb_errors
        sums.loc[label, 'failed'] = nb_combos - nb_matches - nb_errors

    return plot_stacked_bars(
        sums, '# tested e-models', 'Morphology name',
        'Number of tested e-models for each morphology', [BLUE, YELLOW, RED])


def plot_emodels_per_metype(data, final_db):
    """Display result of tested e-model / morphology combinations per me-type.

    Args:
        data: pandas.DataFrame with data on run combos
        final_db: pandas.DataFrame with data on selected combos

    Returns:
        Figure with plot of stacked bars. Simulations that passed, threw an
        error, and failed are colored blue, yellow and red, respectively.
    """
    # Add helper column 'metype'
    def create_metype(x):
        """Create me-type from m-type and e-type"""
        return '%s_%s' % (x['etype'], x['fullmtype'])
    data['metype'] = data.apply(create_metype, axis=1)
    final_db['metype'] = final_db.apply(create_metype, axis=1)

    sums = pandas.DataFrame()
    non_exemplars = data[data['is_exemplar'] == 0]
    for metype in non_exemplars['metype'].unique():
        nb_matches = len(final_db[(final_db['metype'] == metype)])
        nb_errors = len(
            non_exemplars[
                (non_exemplars['metype'] == metype) & (
                    non_exemplars['exception'].notnull())])
        nb_combos = len(non_exemplars[non_exemplars['metype'] == metype])
        sums.loc[metype, 'passed'] = nb_matches
        sums.loc[metype, 'error'] = nb_errors
        sums.loc[metype, 'failed'] = nb_combos - nb_matches - nb_errors

    # Remove helper column 'metype'
    del data['metype']
    del final_db['metype']

    return plot_stacked_bars(
        sums, '# tested (e-model, morphology) combinations', 'me-type',
        'Number of tested (e-model, morphology) combinations per me-type',
        [BLUE, YELLOW, RED], log=True, yticksize=3)


def create_metype(x):
    """Create me-type from m-type and e-type"""
    return '%s_%s' % (x['etype'], x['fullmtype'])


def plot_median_per_metype(combos, passed_median_scores, csv_path):
    """Display result median score per me-type"""

    metype_medians = passed_median_scores.join(combos)[
        ['mtype', 'etype', 'median_score']]

    metype_medians = metype_medians.groupby(
        ['mtype', 'etype']).median().reset_index()

    metype_medians = metype_medians.pivot(
        index='mtype',
        columns='etype',
        values='median_score')

    metype_medians.to_csv(csv_path)
    print('Wrote me-type median scores to %s' % os.path.abspath(csv_path))

    ax = plt.pcolor(metype_medians)
    ax = plt.gca()
    cbar = plt.colorbar()
    cbar.ax.text(
        4, .5, 'median Z score', rotation=270, verticalalignment='center')

    plt.xlabel('e-type')
    plt.ylabel('m-type')

    ax.set_xticks(numpy.arange(metype_medians.shape[1]) + 0.5, minor=False)
    ax.set_yticks(numpy.arange(metype_medians.shape[0]) + 0.5, minor=False)
    ax.set_xticklabels(metype_medians.columns, rotation=90)
    ax.set_yticklabels(metype_medians.index)
    plt.tight_layout()

    plt.title('Median Z scores of me-types')

    return plt.gcf()

# TODO: can this function be split into processing and reporting?


def create_final_db_and_write_report(pdf_filename,
                                     to_skip_features,
                                     to_skip_patterns,
                                     megate_thresholds,
                                     megate_patterns,
                                     skip_repaired_exemplar,
                                     check_opt_scores,
                                     scores,
                                     score_values,
                                     enable_plot_emodels_per_morphology,
                                     output_dir,
                                     select_perc_best,
                                     n_processes=None):
    """Create the final output files and report"""
    ext_neurondb = pandas.DataFrame()

    emodel_infos = None
    megate_passed_all = pandas.DataFrame()
    median_scores = pandas.DataFrame()
    passed_combos = pandas.DataFrame()

    '''
    median_scores = score_values.median(
        axis=1,
        skipna=True).to_frame(name='median_score')
    '''

    with pdf_file(pdf_filename) as pp:
        # Plot input configuration details
        add_plot_to_report(pp, plot_dict, to_skip_features,
                           'Ignored feature patterns')
        add_plot_to_report(pp, plot_dict, megate_thresholds,
                           'MEGating thresholds (last match counts)')

        # Process all the e-models
        emodels = sorted(scores[scores.is_original == 0].emodel.unique())

        emodel_infos = table_processing.process_emodels(emodels,
                                                        scores,
                                                        score_values,
                                                        to_skip_patterns,
                                                        megate_patterns,
                                                        skip_repaired_exemplar,
                                                        check_opt_scores,
                                                        select_perc_best,
                                                        n_processes=n_processes
                                                        )

        print("All emodels processed, generating output files")

        for emodel, emodel_info in emodel_infos.items():
            if emodel_info is not None:
                emodel_ext_neurondb_rows, \
                    megate_scores, emodel_score_values, fullmtypes, \
                    emodel_megate_passed_all, emodel_median_scores, \
                    emodel_passed_combos = \
                    emodel_info
                ext_neurondb = pandas.concat(
                    [ext_neurondb, emodel_ext_neurondb_rows], axis=0
                ).reset_index(drop=True)

                megate_passed_all = pandas.concat(
                    [megate_passed_all, emodel_megate_passed_all], axis=0
                ).reset_index(drop=True)

                median_scores = pandas.concat(
                    [median_scores, emodel_median_scores], axis=0
                ).reset_index(drop=True)

                passed_combos = pandas.concat(
                    [passed_combos, emodel_passed_combos], axis=0
                ).reset_index(drop=True)

                # Reporting per e-model
                add_plot_to_report(
                    pp,
                    plot_morphs_per_feature_for_emodel,
                    emodel,
                    megate_scores,
                    emodel_score_values)
                add_plot_to_report(
                    pp,
                    plot_morphs_per_mtype_for_emodel,
                    emodel,
                    fullmtypes,
                    megate_scores)
                plt.close('all')
            else:
                print('WARNING: no info for emodel %s, skipping !' % emodel)

        # Get median score for every passed combo
        passed_median_scores = median_scores.loc[passed_combos.index]

        extra_data_dir = os.path.join(output_dir, 'extra_data')
        if not os.path.exists(extra_data_dir):
            os.makedirs(extra_data_dir)

        all_median_csv_path = os.path.join(
            extra_data_dir,
            'all_median_scores.csv')
        scores[scores['is_exemplar'] == 0].join(median_scores)[
            ['fullmtype',
             'etype',
             'emodel',
             'median_score']].to_csv(all_median_csv_path)

        passed_median_csv_path = os.path.join(
            extra_data_dir,
            'passed_median_scores.csv')
        scores[
            scores['is_exemplar'] == 0].join(
            passed_median_scores, how='right')[
            ['fullmtype', 'etype', 'emodel', 'median_score']].to_csv(
            passed_median_csv_path)

        metype_median_csv_path = os.path.join(
            extra_data_dir,
            'metype_median_scores.csv')
        add_plot_to_report(
            pp,
            plot_median_per_metype,
            scores,
            passed_median_scores,
            metype_median_csv_path)

        # More reporting
        if enable_plot_emodels_per_morphology:
            add_plot_to_report(pp, plot_emodels_per_morphology, scores,
                               ext_neurondb)
        add_plot_to_report(pp, plot_emodels_per_metype, scores, ext_neurondb)

    return ext_neurondb



"""Test bluepymm main interface"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import shutil
import filecmp

import bluepymm


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DATA_DIR = os.path.join(BASE_DIR, 'examples/simple1')
TMP_DIR = os.path.join(BASE_DIR, 'tmp/main')


def teardown_module():
    """Remove the temporary files."""
    shutil.rmtree(TMP_DIR)


def _verify_emodel_json(filename, output_dir, nb_emodels):
    """Helper function to verify the emodel json file"""
    data_json = os.path.join(output_dir, filename)
    assert os.path.isfile(data_json)
    data = bluepymm.tools.load_json(data_json)
    assert len(data) == nb_emodels
    return data


def _verify_prepare_combos_output(scores_db, emodels_hoc_dir, output_dir,
                                  nb_emodels):
    """Helper function to verify the output of the prepare combos step"""
    # TODO: test database contents
    assert os.path.isfile(scores_db)

    assert os.path.isdir(emodels_hoc_dir)
    hoc_files = os.listdir(emodels_hoc_dir)
    assert len(hoc_files) == nb_emodels
    for hoc_file in hoc_files:
        assert hoc_file.endswith('.hoc')

    _verify_emodel_json('final.json', output_dir, nb_emodels)
    emodel_dirs = _verify_emodel_json('emodel_dirs.json', output_dir,
                                      nb_emodels)
    for emodel in emodel_dirs:
        assert os.path.isdir(emodel_dirs[emodel])


def _verify_run_combos_output(scores_db):
    """Helper function to verify the output of the run combos step"""
    assert os.path.isfile(scores_db)

    # TODO: test database contents
    # Disabled for now, there are absolute paths in db
    """
    import pandas
    import sqlite3

    scores_sqlite_filename = 'output/scores.sqlite'
    exp_scores_sqlite_filename = 'output_expected/scores.sqlite'
    with sqlite3.connect(scores_sqlite_filename) as conn:
        scores = pandas.read_sql('SELECT * FROM scores', conn)

    with sqlite3.connect(exp_scores_sqlite_filename) as conn:
        exp_scores = pandas.read_sql('SELECT * FROM scores', conn)

    if not scores.equals(exp_scores):
        print "Resulting scores db: ", scores
        print "Expected scored db:", exp_scores

    nt.assert_true(scores.equals(exp_scores))
    """


def _verify_select_combos_output(benchmark_dir, output_dir):
    """Helper function to verify output of combination selection"""
    files = ['extneurondb.dat', 'mecombo_emodel.tsv']
    matches = filecmp.cmpfiles(benchmark_dir, output_dir, files)
    if len(matches[0]) != len(files):
        print('Mismatch in files: {}'.format(matches[1]))
    assert len(matches[0]) == len(files)


def _new_prepare_json(original_filename, test_dir):
    """Helper function to prepare new configuration file for prepare_combos."""
    config = bluepymm.tools.load_json(original_filename)
    config['tmp_dir'] = os.path.join(test_dir, 'tmp')
    config['output_dir'] = os.path.join(test_dir, 'output')
    config['scores_db'] = os.path.join(config['output_dir'], 'scores.sqlite')
    config['emodels_hoc_dir'] = os.path.join(config['output_dir'],
                                             'emodels_hoc')
    return bluepymm.tools.write_json(test_dir, original_filename, config)


def _new_run_json(original_filename, test_dir):
    """Helper function to prepare new configuration file for run_combos."""
    config = bluepymm.tools.load_json(original_filename)
    config['output_dir'] = os.path.join(test_dir, 'output')
    config['scores_db'] = os.path.join(config['output_dir'], 'scores.sqlite')
    return bluepymm.tools.write_json(test_dir, original_filename, config)


def _new_select_json(original_filename, test_dir):
    """Helper function to prepare new configuration file for select_combos."""
    config = bluepymm.tools.load_json(original_filename)
    config['scores_db'] = os.path.join(test_dir, 'output', 'scores.sqlite')
    config['pdf_filename'] = os.path.join(test_dir, 'megating.pdf')
    config['output_dir'] = os.path.join(test_dir, 'output')
    return bluepymm.tools.write_json(test_dir, original_filename, config)


def _test_main(test_data_dir, prepare_config_json, run_config_json,
               select_config_json, nb_emodels, test_dir):
    """Helper function to test complete BluePyMM workflow"""

    bluepymm.tools.makedirs(test_dir)

    with bluepymm.tools.cd(test_data_dir):
        # prepare new configuration files based on 'test_dir'
        prepare_config_json = _new_prepare_json(prepare_config_json, test_dir)
        run_config_json = _new_run_json(run_config_json, test_dir)
        select_config_json = _new_select_json(select_config_json, test_dir)

        # prepare combinations
        args_list = ['prepare', prepare_config_json]
        bluepymm.main.run(args_list)

        # verify prepared combinations
        prepare_config = bluepymm.tools.load_json(prepare_config_json)
        _verify_prepare_combos_output(prepare_config['scores_db'],
                                      prepare_config['emodels_hoc_dir'],
                                      prepare_config['output_dir'], nb_emodels)

        # run combinations
        args_list = ['run', run_config_json]
        bluepymm.main.run(args_list)

        # verify run combinations
        run_config = bluepymm.tools.load_json(run_config_json)
        _verify_run_combos_output(run_config['scores_db'])

        # select combinations
        args_list = ['select', select_config_json]
        bluepymm.main.run(args_list)

        # test selection output
        select_config = bluepymm.tools.load_json(select_config_json)
        _verify_select_combos_output('output_megate_expected',
                                     select_config['output_dir'])


def test_main_from_dir():
    """bluepymm.main: test full BluePyMM workflow with plain directory input
    based on example simple1"""
    prepare_config_json = 'simple1_conf_prepare.json'
    run_config_json = 'simple1_conf_run.json'
    select_config_json = 'simple1_conf_select.json'
    nb_emodels = 2
    test_dir = os.path.join(TMP_DIR, 'test_main_from_dir')

    _test_main(TEST_DATA_DIR, prepare_config_json, run_config_json,
               select_config_json, nb_emodels, test_dir)


def test_main_from_git_repo():
    """bluepymm.main: test full BluePyMM workflow with git repo input
    based on example simple1"""
    prepare_config_json = 'simple1_conf_prepare_git.json'
    run_config_json = 'simple1_conf_run.json'
    select_config_json = 'simple1_conf_select.json'
    nb_emodels = 2
    test_dir = os.path.join(TMP_DIR, 'test_main_from_git_repo')

    _test_main(TEST_DATA_DIR, prepare_config_json, run_config_json,
               select_config_json, nb_emodels, test_dir)



"""Test table processing"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import json
import pandas
import os
import re
from string import digits

import pytest

from bluepymm.select_combos import table_processing
from bluepymm.select_combos import process_megate_config as proc_config
from bluepymm import tools


BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TEST_DIR = os.path.join(BASE_DIR, 'examples/simple1')


@pytest.mark.unit
def test_convert_extra_values():
    """select_combos.table_processing: test convert_extra_values"""

    value = 42
    # dict grows with given field
    for field in ['threshold_current', 'holding_current']:
        data = {'extra_values': json.dumps({field: value})}
        ret = table_processing.convert_extra_values(data)
        assert ret[field] == value

    # dict does not change
    for field in ['random']:
        data = {'extra_values': json.dumps({field: value})}
        ret = table_processing.convert_extra_values(data)
        assert ret == data


@pytest.mark.unit
def test_row_threshold_transform():
    """select_combos.table_processing: test row_threshold_transform"""

    # megate_feature_threshold is None
    row = {'emodel': 'test1', 'fullmtype': 'test2', 'etype': 'test3',
           'megate_feature_threshold': None}
    patterns = [
        {'megate_feature_threshold':
         {'megate_threshold': 5, 'features': proc_config.join_regex(
             ['.*'])}, 'emodel': proc_config.join_regex(['test1']),
         'fullmtype': proc_config.join_regex(['test2']), 'etype':
         proc_config.join_regex(['test3'])}]
    ret = table_processing.row_threshold_transform(row, patterns)
    assert (
        ret["megate_feature_threshold"][0]
        == patterns[0]["megate_feature_threshold"]
    )

    # megate_feature_threshold is not None
    row = {'emodel': 'test1', 'fullmtype': 'test2', 'etype': 'test3',
           'megate_feature_threshold': [
               patterns[0]['megate_feature_threshold']]}
    ret = table_processing.row_threshold_transform(row, patterns)
    expected_list = [patterns[0]['megate_feature_threshold'],
                     patterns[0]['megate_feature_threshold']]
    assert ret['megate_feature_threshold'] == expected_list


@pytest.mark.unit
def test_check_opt_scores():
    """select_combos.table_processing: test check_opt_scores"""
    # everything OK
    emodel = 'emodel1'
    scores_dict = {'emodel': [emodel], 'is_exemplar': [1], 'is_repaired': [0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 20.0})],
                   'scores': [json.dumps({'Step1.SpikeCount': 20.0})],
                   }
    scores = pandas.DataFrame(scores_dict)
    table_processing.check_opt_scores(emodel, scores)

    # different keys
    scores_dict = {'emodel': [emodel], 'is_exemplar': [1], 'is_repaired': [0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 20.0})],
                   'scores': [json.dumps({'Step2.SpikeCount': 20.0})],
                   }
    scores = pandas.DataFrame(scores_dict)
    with pytest.raises(Exception):
        table_processing.check_opt_scores(emodel, scores)

    # different score values
    scores_dict = {'emodel': [emodel], 'is_exemplar': [1], 'is_repaired': [0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 20.0})],
                   'scores': [json.dumps({'Step1.SpikeCount': 10.0})],
                   }
    scores = pandas.DataFrame(scores_dict)
    with pytest.raises(Exception):
        table_processing.check_opt_scores(emodel, scores)


@pytest.mark.unit
def test_process_emodel():
    # input parameters
    emodel = 'emodel1'
    mtypes = ['mtype1', 'mtype2']
    scores_dict = {'emodel': [emodel, emodel],
                   'is_exemplar': [1, 0], 'is_repaired': [1, 0],
                   'is_original': [0, 0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                                  json.dumps({'Step1.SpikeCount': 2.0})],
                   'scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                              json.dumps({'Step1.SpikeCount': 2.0})],
                   'etype': ['etype1', 'etype2'],
                   'fullmtype': mtypes,
                   'mtype': mtypes,
                   'extra_values': [json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0}),
                                    json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0})],
                   'morph_name': ['morph1', 'morph1'],
                   'layer': ['layer_1', 'layer_1']
                   }
    scores = pandas.DataFrame(scores_dict)
    score_values_dict = {'Step1.SpikeCount': [2.0, 2.0]}
    score_values = pandas.DataFrame(score_values_dict)
    to_skip_patterns = []
    regex_all = re.compile('.*')
    megate_patterns = [{'megate_feature_threshold': {'megate_threshold': 5,
                                                     'features': regex_all},
                        'emodel': regex_all, 'fullmtype': regex_all,
                        'etype': regex_all}]

    skip_repaired_exemplar = False
    enable_check_opt_scores = True

    select_best_perc = None

    emodel, emodel_info = table_processing.process_emodel(
        (emodel, scores, score_values, to_skip_patterns, megate_patterns,
         skip_repaired_exemplar, enable_check_opt_scores, select_best_perc))

    (emodel_ext_neurondb, emodel_megate_pass, emodel_score_values,
     mtypes, emodel_megate_passed_all, emodel_median_scores,
     passed_combos) = emodel_info

    # expected results
    columns = ['morph_name', 'layer', 'fullmtype', 'etype', 'emodel',
               'combo_name', 'threshold_current', 'holding_current']
    db_dict = {'morph_name': 'morph1',
               'layer': 'layer_1',
               'fullmtype': 'mtype2',
               'etype': 'etype2',
               'emodel': 'emodel1',
               'threshold_current': 0.0,
               'holding_current': 0.0,
               'combo_name': 'emodel1_mtype2_layer_1_morph1'
               }
    exp_db = pandas.DataFrame(data=db_dict, columns=columns, index=[1])

    columns = ['Step1.SpikeCount', 'Passed all']
    megate_scores_dict = {'Step1.SpikeCount': True, 'Passed all': True}
    exp_megate_scores = pandas.DataFrame(data=megate_scores_dict,
                                         columns=columns, index=[1])

    exp_score_values = pandas.DataFrame({'Step1.SpikeCount': 2.0}, index=[1])

    exp_mtypes = pandas.Series('mtype2', index=[1], name='fullmtype')

    # verify results
    pandas.testing.assert_frame_equal(emodel_ext_neurondb, exp_db)
    pandas.testing.assert_frame_equal(
        emodel_megate_pass, exp_megate_scores)
    pandas.testing.assert_frame_equal(
        emodel_score_values, exp_score_values)
    pandas.testing.assert_series_equal(mtypes, exp_mtypes)


@pytest.mark.unit
def test_process_emodel_no_exemplars():
    # input parameters
    emodel = 'emodel1'
    mtypes = ['mtype1', 'mtype2']
    scores_dict = {'emodel': [emodel, emodel],
                   'is_exemplar': [0, 0], 'is_repaired': [1, 0],
                   'is_original': [1, 0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                                  json.dumps({'Step1.SpikeCount': 2.0})],
                   'scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                              json.dumps({'Step1.SpikeCount': 2.0})],
                   'etype': ['etype1', 'etype2'],
                   'fullmtype': mtypes,
                   'mtype': mtypes,
                   'extra_values': [json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0}),
                                    json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0})],
                   'morph_name': ['morph1', 'morph1'],
                   'layer': ['layer_1', 'layer_1']
                   }
    scores = pandas.DataFrame(scores_dict)
    score_values_dict = {'Step1.SpikeCount': [2.0, 2.0]}
    score_values = pandas.DataFrame(score_values_dict)
    to_skip_patterns = []
    regex_all = re.compile('.*')
    megate_patterns = [{'megate_feature_threshold': {'megate_threshold': 5,
                                                     'features': regex_all},
                        'emodel': regex_all, 'fullmtype': regex_all,
                        'etype': regex_all}]

    skip_repaired_exemplar = False
    enable_check_opt_scores = True

    select_best_perc = None

    # run function
    ret = table_processing.process_emodel((
        emodel, scores, score_values, to_skip_patterns, megate_patterns,
        skip_repaired_exemplar, enable_check_opt_scores, select_best_perc))

    # verify results
    assert ret is None


@pytest.mark.unit
def test_process_emodel_too_many_exemplars():
    # input parameters
    emodel = 'emodel1'
    mtypes = ['mtype1', 'mtype2']
    scores_dict = {'emodel': [emodel, emodel],
                   'is_exemplar': [1, 1], 'is_repaired': [1, 1],
                   'is_original': [0, 0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                                  json.dumps({'Step1.SpikeCount': 2.0})],
                   'scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                              json.dumps({'Step1.SpikeCount': 2.0})],
                   'etype': ['etype1', 'etype2'],
                   'fullmtype': mtypes,
                   'mtype': mtypes,
                   'extra_values': [json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0}),
                                    json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0})],
                   'morph_name': ['morph1', 'morph1'],
                   'layer': ['layer_1', 'layer_1']
                   }
    scores = pandas.DataFrame(scores_dict)
    score_values_dict = {'Step1.SpikeCount': [2.0, 2.0]}
    score_values = pandas.DataFrame(score_values_dict)
    to_skip_patterns = []
    regex_all = re.compile('.*')
    megate_patterns = [{'megate_feature_threshold': {'megate_threshold': 5,
                                                     'features': regex_all},
                        'emodel': regex_all, 'fullmtype': regex_all,
                        'etype': regex_all}]

    skip_repaired_exemplar = False
    enable_check_opt_scores = True

    # run function
    with pytest.raises(Exception):
        table_processing.process_emodel(
            emodel,
            scores,
            score_values,
            to_skip_patterns,
            megate_patterns,
            skip_repaired_exemplar,
            enable_check_opt_scores,
        )


@pytest.mark.unit
def test_process_emodel_no_released_morphologies():
    # input parameters
    emodel = 'emodel1'
    mtypes = ['mtype1', 'mtype2']
    scores_dict = {'emodel': [emodel, emodel],
                   'is_exemplar': [1, 1], 'is_repaired': [1, 0],
                   'is_original': [1, 0],
                   'opt_scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                                  json.dumps({'Step1.SpikeCount': 2.0})],
                   'scores': [json.dumps({'Step1.SpikeCount': 2.0}),
                              json.dumps({'Step1.SpikeCount': 2.0})],
                   'etype': ['etype1', 'etype2'],
                   'fullmtype': mtypes,
                   'mtype': mtypes,
                   'extra_values': [json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0}),
                                    json.dumps({'threshold_current': 0.0,
                                                'holding_current': 0.0})],
                   'morph_name': ['morph1', 'morph1'],
                   'layer': ['layer_1', 'layer_1']
                   }
    scores = pandas.DataFrame(scores_dict)
    score_values_dict = {'Step1.SpikeCount': [2.0, 2.0]}
    score_values = pandas.DataFrame(score_values_dict)
    to_skip_patterns = []
    regex_all = re.compile('.*')
    megate_patterns = [{'megate_feature_threshold': {'megate_threshold': 5,
                                                     'features': regex_all},
                        'emodel': regex_all, 'fullmtype': regex_all,
                        'etype': regex_all}]

    skip_repaired_exemplar = True
    enable_check_opt_scores = True

    select_best_perc = None
    # run function
    ret = table_processing.process_emodel((
        emodel, scores, score_values, to_skip_patterns, megate_patterns,
        skip_repaired_exemplar, enable_check_opt_scores, select_best_perc))

    # verify results
    assert ret == ('emodel1', None)


@pytest.mark.unit
def test_process_combo_name():
    """select_combos.table_processing: test process_combo_name"""
    combo_names = [
        'test', '123test-test',
        'testtesttesttesttesttesttesttesttesttesttesttesttesttesttesttest']
    data = pandas.DataFrame(combo_names, columns=['combo_name'])
    log_filename = os.path.join(TEST_DIR, 'log.csv')
    # make sure the file does not exist yet
    if os.path.isfile(log_filename):
        os.remove(log_filename)

    # compose expected NEURON-compliant names
    expected_names = []
    assert tools.check_compliance_with_neuron(combo_names[0])
    expected_names.append(combo_names[0])
    assert not tools.check_compliance_with_neuron(combo_names[1])
    name = combo_names[1].lstrip(digits).replace('-', '_')
    assert tools.check_compliance_with_neuron(name)
    expected_names.append(name)
    assert not tools.check_compliance_with_neuron(combo_names[2])
    name = tools.get_neuron_compliant_template_name(combo_names[2])
    assert tools.check_compliance_with_neuron(name)
    expected_names.append(name)

    expected_df = pandas.DataFrame(expected_names, columns=['combo_name'])

    table_processing.process_combo_name(data, log_filename)
    pandas.testing.assert_frame_equal(data, expected_df)
    assert os.path.isfile(log_filename)
    # clear output
    if os.path.isfile(log_filename):
        os.remove(log_filename)



"""Functions to process tables produced by BluePyMM."""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0914, C0325, W0640
# pylama: ignore=E402

import math
import json
import pandas
import multiprocessing

from bluepymm import tools


def _row_transform(row, exemplar_row, to_skip_patterns,
                   skip_repaired_exemplar):
    """Transform row values (scores) to booleans, where True means that a
    feature did not exceed the corresponding feature threshold, or can be
    ignored.
    """

    for column in row.index[1:]:
        for megate_feature_threshold in row['megate_feature_threshold']:
            if megate_feature_threshold['features'].match(column):
                megate_threshold = megate_feature_threshold['megate_threshold']

        # transform score
        if skip_repaired_exemplar:
            row[column] = row[column] <= megate_threshold
        else:
            row[column] = row[column] <= max(
                megate_threshold, megate_threshold * exemplar_row[column])

    return row


def convert_extra_values(row):
    """Extract 'threshold_current' and 'holding_current' information from key
    'extra_values' and convert to new (key, value)-pairs in given row data.

    Args:
        row: contains key 'extra_values', with string value

    Returns:
        row, with extra keys 'threshold_current' and/or 'holding_current', and
        associated value, as extracted from row['extra_values']
    """

    extra_values_str = row['extra_values']

    if extra_values_str is not None:
        extra_values = json.loads(extra_values_str)
        if extra_values:
            for field in ['threshold_current', 'holding_current']:
                if field in extra_values:
                    row[field] = extra_values[field]
    return row


def row_threshold_transform(row, megate_patterns):
    """Transform threshold row based on me-gate rule: add matching me-gate
    patterns to row data.

    Args:
        row: has keys 'emodel', 'fullmtype', 'etype', and
            'megate_feature_threshold'
        megate_patterns: a list of megate patterns

    Returns:
        Processed row data: for all me-gate patterns that match the row data,
        the corresponding megate feature threshold is appended to
        row['megate_feature_threshold'].
    """

    emodel = row['emodel']
    fullmtype = row['fullmtype']
    etype = row['etype']

    for pattern_dict in megate_patterns:
        if (pattern_dict['emodel'].match(emodel) and
           pattern_dict['fullmtype'].match(fullmtype) and
           pattern_dict['etype'].match(etype)):
            if row['megate_feature_threshold'] is None:
                row['megate_feature_threshold'] = []
            row['megate_feature_threshold'].append(pattern_dict[
                'megate_feature_threshold'])

    return row


def check_opt_scores(emodel, scores):
    """Check if opt_scores match with unrepaired exemplar runs.

    Args:
        emodel: e-model name
        scores: pandas.DataFrame with scores

    Raises:
        Exception:
            - if the keys of the opt_scores do not match the unrepaired
              exemplar runs,
            - if the scores values of the opt_scores do not match the scores
              of the unrepaired exemplar runs.
    """
    test_rows = scores[(scores.emodel == emodel) &
                       (scores.is_exemplar == 1) &
                       (scores.is_repaired == 0)]

    for _, row in test_rows.iterrows():
        opt_score = json.loads(row['opt_scores'])
        bluepymm_score = json.loads(row['scores'])

        if bluepymm_score is not None:
            if sorted(opt_score.keys()) != sorted(bluepymm_score.keys()):
                raise Exception(
                    'Difference detected in score keys between optimisation'
                    'score and score calculated by bluepymm for emodel %s !:'
                    '\n%s\n%s' %
                    (emodel, opt_score, bluepymm_score))

            for feature in opt_score:
                if opt_score[feature] != bluepymm_score[feature]:
                    raise Exception(
                        'Difference detected in optimisation score and score '
                        'calculated by bluepymm for emodel %s !:\n%s\n%s' %
                        (emodel, opt_score, bluepymm_score))


def _apply_megating(emodel_mtype_etype_thresholds, emodel_score_values,
                    exemplar_row, to_skip_patterns, skip_repaired_exemplar):
    """Compare score values to applicable feature thresholds."""

    # Add a column with the thresholds
    emodel_score_values_thresholds = pandas.concat(
        [emodel_mtype_etype_thresholds['megate_feature_threshold'],
         emodel_score_values], axis=1)

    # Apply the thresholds
    # Creates a table show which columns (objectives) pass for each combo

    non_skipped_columns = [
        column
        for column in
        emodel_score_values_thresholds.columns
        if not any(pattern.match(column) for pattern in to_skip_patterns)]

    emodel_megate_pass = \
        emodel_score_values_thresholds[non_skipped_columns].apply(
            _row_transform, args=(
                exemplar_row,
                to_skip_patterns,
                skip_repaired_exemplar),
            axis=1)

    # Remove the threshold column
    del emodel_megate_pass['megate_feature_threshold']

    # Detect which rows (combos) pass in all columns
    emodel_megate_pass['Passed all'] = emodel_megate_pass.all(axis=1)
    return emodel_megate_pass


def _create_extneurondb_rows(selected_combinations):
    """Prepare rows for database based on selected combinations."""
    # 1. select relevant columns from db with successful combinations
    emodel_ext_neurondb = selected_combinations.loc[:, ('morph_name',
                                                        'layer',
                                                        'fullmtype',
                                                        'etype',
                                                        'emodel',
                                                        'extra_values')].copy()

    # 2. create additional columns: combo_name, threshold current, and
    #    holding current
    if len(emodel_ext_neurondb) > 0:
        emodel_ext_neurondb['combo_name'] = emodel_ext_neurondb.apply(
            lambda x: '%s_%s_%s_%s' %
            (x['emodel'], x['fullmtype'], x['layer'], x['morph_name']), axis=1)

        emodel_ext_neurondb['threshold_current'] = None
        emodel_ext_neurondb['holding_current'] = None
        emodel_ext_neurondb = emodel_ext_neurondb.apply(
            convert_extra_values, axis=1)
        del emodel_ext_neurondb['extra_values']

    return emodel_ext_neurondb


def select_passed_combos(
        emodel,
        emodel_combos,
        emodel_megate_pass,
        emodel_megate_scores,
        select_perc_best=None):
    """Select which combos pass"""

    passed_indices = pandas.DataFrame()

    if select_perc_best is not None:
        metype_inds = emodel_combos['etype'] + emodel_combos['fullmtype']
        for metype in metype_inds.unique():
            metype_scores = emodel_megate_scores.loc[
                metype_inds[
                    metype_inds.values == metype].index]
            metype_scores_nonan = metype_scores.dropna(axis=0)
            metype_scores_nonan_no250 = metype_scores_nonan[
                metype_scores_nonan['median_score'] < 250.0]
            # metype_scores_nonan = metype_scores
            metype_scores_nonan_no250_sorted = \
                metype_scores_nonan_no250.sort_values(
                    'median_score')
            n_of_combos = len(metype_scores_nonan_no250_sorted.index)
            n_of_best = int(math.ceil(select_perc_best * n_of_combos))

            passed_indices = passed_indices.append(
                metype_scores_nonan_no250_sorted.head(n_of_best))

            if len(passed_indices) == 0:
                print(
                    'WARNING: no combos for me-type %s in emodel %s' %
                    (metype, emodel))
    else:
        passed_indices = \
                emodel_megate_pass[
                emodel_megate_pass['Passed all'] == True]  # NOQA

    return emodel_combos.loc[passed_indices.index]


def calc_median_scores(emodel_score_values, to_skip_patterns):
    """Calculate scores for every me-combo"""

    columns = emodel_score_values.columns

    non_skipped_columns = [
        column
        for column in
        columns
        if not any(pattern.match(column) for pattern in to_skip_patterns)]

    emodel_median_scores = emodel_score_values[non_skipped_columns].median(
        axis=1, skipna=True).to_frame('median_score')

    return emodel_median_scores


def process_emodels(emodels,
                    scores,
                    score_values,
                    to_skip_patterns,
                    megate_patterns,
                    skip_repaired_exemplar,
                    enable_check_opt_scores,
                    select_perc_best,
                    n_processes=None):

    arg_list = [(emodel,
                 scores,
                 score_values,
                 to_skip_patterns,
                 megate_patterns,
                 skip_repaired_exemplar,
                 enable_check_opt_scores,
                 select_perc_best) for emodel in emodels]

    emodel_infos = {}

    if n_processes == 1:
        for args in arg_list:
            emodel, emodel_info = process_emodel(args)
            emodel_infos[emodel] = emodel_info
    else:
        print('Parallelising selection processing of e-models')
        pool = multiprocessing.Pool(maxtasksperchild=1, processes=n_processes)
        for emodel, emodel_info in pool.imap(process_emodel, arg_list,
                                             chunksize=1):
            print('Received processed info from e-model %s' % emodel)
            emodel_infos[emodel] = emodel_info

        pool.terminate()
        pool.join()

    return emodel_infos


def process_emodel(args):
    """Process scores and score values for indicated e-model and return data
    on the e-model performance as well as the selected combinations.

    Args:
        emodel: e-model name
        combos: pandas.DataFrame with combo data
        score_values: pandas.DataFrame with score values
        to_skip_patterns: list of compiled regular expressions
        megate_patterns: list of dictionaries with megate patterns
        skip_repaired_exemplar: boolean
        enable_check_opt_scores: boolean

    Returns:
        4-tuple with megate results for the e-model:
        - emodel_ext_neurondb: pandas.DataFrame with database rows
        - emodel_megate_pass: pandas.DataFrame with megate fail/success
        - emodel_score_values: pandas.DataFrame with score values
        - mtypes: pandas.DataFrame with tested m-types

        None:
        - if boolean skip_repaired_exemplar is set to False,
        and no repaired exemplars are available
        - if the e-model was not run on any released morphology

    Raises:
        Exception, skip_repaired_exemplar is set to False and more than one
        exemplars are found.
    """
    emodel, combos, score_values, to_skip_patterns, megate_patterns, \
        skip_repaired_exemplar, enable_check_opt_scores, select_perc_best \
        = args

    print('Processing e-model %s' % emodel)

    # check if opt_scores match with unrepaired exemplar runs
    if enable_check_opt_scores:
        check_opt_scores(emodel, combos)

    # if applicable, skip exemplar rows from combos and score values
    exemplar_row = None
    if not skip_repaired_exemplar:
        exemplar_morph = combos[combos.emodel == emodel].morph_name.values[0]
        exemplar_score_values = score_values[
            (combos.emodel == emodel) &
            (combos.is_exemplar == 1) &
            (combos.is_repaired == 1) &
            (combos.is_original == 0) &
            (combos.morph_name == exemplar_morph)]

        if len(exemplar_score_values) > 1:
            raise Exception('Too many exemplars found for e-model %s: %s' %
                            (emodel, exemplar_score_values))

        exemplar_score_values = exemplar_score_values.head(1).copy()
        exemplar_score_values.dropna(axis=1, how='all', inplace=True)

        if len(exemplar_score_values) == 0:
            print('Skipping e-model %s: no repaired exemplars' % emodel)
            return

        exemplar_row = exemplar_score_values.iloc[0].to_dict()

    # identify relevant me-gate feature thresholds for each row
    emodel_mtype_etypes = combos[(combos.emodel == emodel) &
                                 (combos.is_exemplar == 0)].copy()
    if len(emodel_mtype_etypes) == 0:
        print('Skipping e-model %s: was not run on any released morphology'
              % emodel)
        return (emodel, None)

    emodel_mtype_etype_thresholds = emodel_mtype_etypes.loc[
        :, ['emodel', 'fullmtype', 'etype']]
    emodel_mtype_etype_thresholds['megate_feature_threshold'] = None

    print('Getting megating thresholds for emodel %s' % emodel)
    emodel_mtype_etype_thresholds = emodel_mtype_etype_thresholds.apply(
        lambda row: row_threshold_transform(row, megate_patterns),
        axis=1)

    # select score values relevant to this e-model
    emodel_score_values = score_values[(combos.emodel == emodel) &
                                       (combos.is_exemplar == 0)].copy()
    emodel_score_values.dropna(axis=1, how='all', inplace=True)

    print('Applying megating to emodel %s' % emodel)
    # me-gating: compare score values to applicable feature thresholds
    emodel_megate_pass = _apply_megating(
        emodel_mtype_etype_thresholds,
        emodel_score_values,
        exemplar_row,
        to_skip_patterns,
        skip_repaired_exemplar)

    print('Calculating median scores for emodel %s' % emodel)

    emodel_median_scores = calc_median_scores(
        emodel_score_values,
        to_skip_patterns)

    emodel_combos = combos[(combos.emodel == emodel) &
                           (combos.is_exemplar == 0)].copy()

    # identify combinations that passed the me-gating step
    passed_combos = select_passed_combos(
        emodel,
        emodel_combos,
        emodel_megate_pass,
        emodel_median_scores,
        select_perc_best)
    emodel_megate_passed_all = emodel_megate_pass[['Passed all']]

    if len(passed_combos[passed_combos['emodel'] != emodel]) > 0:
        raise Exception('Something went wrong during row indexing in megating')

    # prepare database rows for this e-model
    emodel_ext_neurondb = _create_extneurondb_rows(passed_combos)

    # identify m-types that were tested for this e-model
    mtypes = combos[(combos.emodel == emodel) &
                    (combos.is_exemplar == 0)].loc[:, 'fullmtype']

    return emodel, (emodel_ext_neurondb, emodel_megate_pass,
                    emodel_score_values, mtypes, emodel_megate_passed_all,
                    emodel_median_scores, passed_combos)


def process_combo_name(data, log_filename):
    """Make value corresponding to key 'combo_name' compliant with NEURON rules
    for template names. A log file is written out in csv format.

    Args:
        data: pandas.DataFrame with key 'combo_name'
        log_filename: path to log file
    """
    log_data = pandas.DataFrame()
    log_data['original_combo_name'] = data['combo_name'].copy()

    data['combo_name'] = data.apply(
        lambda x: tools.get_neuron_compliant_template_name(x['combo_name']),
        axis=1)

    log_data['neuron_compliant_combo_name'] = data['combo_name'].copy()
    log_data.to_csv(log_filename, index=False)



"""Process megate configuration file."""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0914, C0325, W0640


import re


def join_regex(list_regex):
    """Create regular expresssion that matches one of a given list of regular
    expressions."""

    list_regex_dollar = []
    for regex in list_regex:
        if regex[-1] != '$':
            regex_dollar = regex + '$'

        list_regex_dollar.append(regex_dollar)

    return re.compile('(' + ')|('.join(list_regex_dollar) + ')')


def read_to_skip_features(conf_dict):
    """Parse features to skip from configuration and return list of compiled
    regular expressions.

    Args:
        conf_dict: dictionary, value of conf_dict['to_skip_features'] is
            processed if available.

    Returns:
        A tuple (<list_of_compiled_reg_exprs>, <conf_dict['to_skip_features']>)
    """

    to_skip_features = conf_dict.get('to_skip_features', [])

    return [re.compile(feature_str)
            for feature_str in to_skip_features], to_skip_features


def read_megate_thresholds(conf_dict):
    """Parse megate thresholds from configuraiton and return list of compiled
    regular expressions.

    Args:
        conf_dict: dictionary, value of conf_dict['megate_thresholds'] is
            processed if available.

    Returns:
        A tuple (<list_of_dicts>, <conf_dict['megate_thresholds']>)"""

    megate_thresholds = conf_dict.get('megate_thresholds', [])

    megate_patterns = []
    for megate_threshold_dict in megate_thresholds:
        for key in megate_threshold_dict:
            if key not in {'emodel', 'etype', 'fullmtype',
                           'features', 'megate_threshold'}:
                raise ValueError(
                    'Invalid key in megate thresholds: %s in %s' %
                    (key, megate_threshold_dict))

        megate_pattern = {}
        megate_pattern['megate_feature_threshold'] = {
            'megate_threshold': megate_threshold_dict['megate_threshold'],
            'features': join_regex(megate_threshold_dict['features'])
        }
        for key in ['emodel', 'fullmtype', 'etype']:
            if key in megate_threshold_dict:
                megate_pattern[key] = join_regex(megate_threshold_dict[key])
            else:
                megate_pattern[key] = re.compile('.*$')

        megate_patterns.append(megate_pattern)

    return megate_patterns, megate_thresholds



"""Test parse_files"""


"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os

import pytest

import pandas
import xml.etree.ElementTree as ET

from bluepymm.prepare_combos import parse_files

BASE_DIR = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.unit
def test_read_circuitmvd3():
    """bluepymm.prepare_combos.parse_files: test reading a circuit.mvd3"""

    for cmvd3_fn in ['circuit_strlayers.mvd3', 'circuit_intlayers.mvd3']:
        cmvd3_path = os.path.join(
            BASE_DIR, 'examples/cmvd3a', cmvd3_fn)

        cmvd3_content = parse_files.read_circuitmvd3(cmvd3_path)

        expected_layers = ['5', '2'] * 5
        expected_mtypes = ['L5_TPC', 'L2_TPC'] * 5
        expected_etypes = ['cADpyr'] * 10
        expected_morphnames = []

        for index in range(1, 6):
            expected_morphnames += ['M%d_C060114A5' %
                                    index, 'M%d_mtC191200B_idA' % index]
        assert list(cmvd3_content['layer'].values) == expected_layers
        assert list(cmvd3_content['fullmtype'].values) == expected_mtypes
        assert list(cmvd3_content['etype'].values) == expected_etypes
        assert list(cmvd3_content['morph_name'].values) == expected_morphnames


@pytest.mark.unit
def test_verify_no_zero_percentage_no_zero():
    """bluepymm.prepare_combos.parse_files: test nonzero perc in recipe"""
    tree_string = """
        <tree>
            <element percentage="10.0"/>
            <element percentage="20.0"/>
        </tree>
        """
    data = ET.fromstring(tree_string)
    throws_exception = False
    children = list(data)
    try:
        ret = parse_files.verify_no_zero_percentage(children)
        assert ret
    except ValueError:
        throws_exception = True
    assert not throws_exception


@pytest.mark.unit
def test_verify_no_zero_percentage_zero():
    """bluepymm.prepare_combos.parse_files: test zero perc in recipe"""
    tree_string = """
        <tree>
            <element percentage="0.0"/>
            <element percentage="20.0"/>
        </tree>
        """
    data = ET.fromstring(tree_string)
    children = list(data)
    with pytest.raises(ValueError):
        parse_files.verify_no_zero_percentage(children)


@pytest.mark.unit
def test_read_recipe_records():
    """bluepymm.prepare_combos.parse_files: test read_recipe_records.
    """
    tree_string = """
        <Recipe>
            <NeuronTypes>
                <Layer id="1" percentage="1.037315">
                  <StructuralType id="mtype1" percentage="17.7304964539">
                    <ElectroType id="etype1" percentage="33.33"/>
                    <ElectroType id="etype2" percentage="66.67"/>
                  </StructuralType>
                  <StructuralType id="mtype2" percentage="20.5673758865">
                    <ElectroType id="etype1" percentage="10.00"/>
                  </StructuralType>
                </Layer>

                <Layer id="two" percentage="10.832282">
                    <StructuralType id="mtype1" percentage="67.7948717949">
                        <ElectroType id="etype2" percentage="100" />
                    </StructuralType>
                </Layer>
            </NeuronTypes>
        </Recipe>
    """
    expected_records = [("1", "mtype1", "etype1"),
                        ("1", "mtype1", "etype2"),
                        ("1", "mtype2", "etype1"),
                        ("two", "mtype1", "etype2"), ]
    recipe_tree = ET.fromstring(tree_string)
    records = list(parse_files.read_recipe_records(recipe_tree))
    assert records == expected_records


@pytest.mark.unit
def test_read_mm_recipe_xml():
    """bluepymm.prepare_combos.parse_files: test read_mm_recipe with an xml
    recipe from test example "simple1".
    """
    recipe_filename = os.path.join(
        BASE_DIR,
        'examples/simple1/data/simple1_recipe.xml')
    expected_records = [("1", "mtype1", "etype1"),
                        ("1", "mtype1", "etype2"),
                        ("1", "mtype2", "etype1"),
                        ("2", "mtype1", "etype2"), ]
    expected_df = pandas.DataFrame(expected_records,
                                   columns=["layer", "fullmtype", "etype"])
    df = parse_files.read_mm_recipe(recipe_filename)
    pandas.testing.assert_frame_equal(df, expected_df)


@pytest.mark.unit
def test_read_mm_recipe_yaml():
    """bluepymm.prepare_combos.parse_files: test read_mm_recipe with a yaml
    recipe from test example "simple1".
    """
    recipe_filename = os.path.join(
        BASE_DIR,
        'examples/simple1/data/simple1_recipe.yaml')
    expected_records = [("1", "mtype1", "etype1"),
                        ("1", "mtype1", "etype2"),
                        ("1", "mtype2", "etype1"),
                        ("2", "mtype1", "etype2"), ]
    expected_df = pandas.DataFrame(expected_records,
                                   columns=["layer", "fullmtype", "etype"],
                                   )
    expected_df = expected_df.sort_values(by=['etype', 'fullmtype', 'layer'])
    expected_df = expected_df.reset_index(drop=True)

    df = parse_files.read_mm_recipe(recipe_filename)
    df = df.sort_values(by=['etype', 'fullmtype', 'layer'])
    df = df.reset_index(drop=True)

    pandas.testing.assert_frame_equal(df, expected_df, check_dtype=False)


@pytest.mark.unit
def test_read_morph_records():
    """bluepymm.prepare_combos.parse_files: test read_morph_records.
    """
    tree_string = """
        <neurondb>
            <listing>
                <morphology>
                    <name>morph1</name>
                    <mtype>mtype1</mtype>
                    <msubtype />
                    <layer>1</layer>
                </morphology>
                <morphology>
                    <name>morph2</name>
                    <mtype>mtype2</mtype>
                    <msubtype>subtype2</msubtype>
                    <layer>layer2</layer>
                </morphology>
            </listing>
        </neurondb>
    """
    expected_records = [("morph1", "mtype1", "mtype1", "", "1"),
                        ("morph2", "mtype2:subtype2", "mtype2", "subtype2",
                         "layer2")]
    morph_tree = ET.fromstring(tree_string)
    records = list(parse_files.read_morph_records(morph_tree))
    assert records == expected_records


@pytest.mark.unit
def test_read_mtype_morph_map():
    """bluepymm.prepare_combos.parse_files: test read_mtype_morph_map with
    morphology database from test example "simple1".
    """
    neurondb_filename = os.path.join(
        BASE_DIR,
        "examples/simple1/data/morphs/neuronDB.xml")
    expected_records = [("morph1", "mtype1", "mtype1", "", "1"),
                        ("morph2", "mtype2", "mtype2", "", "1")]
    expected_df = pandas.DataFrame(expected_records,
                                   columns=["morph_name", "fullmtype", "mtype",
                                            "submtype", "layer"])
    df = parse_files.read_mtype_morph_map(neurondb_filename)
    pandas.testing.assert_frame_equal(df, expected_df)


def _test_convert_emodel_etype_map(emodel_etype_map, fullmtypes, etypes,
                                   layers):
    """test convert emodel etype map"""
    df = parse_files.convert_emodel_etype_map(emodel_etype_map, fullmtypes,
                                              etypes)
    assert len(df) == len(fullmtypes) * len(layers)
    assert df["original_emodel"].unique() == ["emodel1"]
    assert df["emodel"].unique() == ["emodel2"]

    for mtype in fullmtypes:
        df_layers = list(df[df["fullmtype"] == mtype]["layer"].unique())
        assert df_layers == layers

    assert len(df.drop_duplicates().index) == len(fullmtypes) * len(layers)


@pytest.mark.unit
def test_convert_emodel_etype_map_no_regex():
    """prepare_combos.parse_files: test emodel etype map convert w/ regex"""
    layers = ["layer1", "2"]
    emodel_etype_map = {"emodel1": {"mm_recipe": "emodel2",
                                    "layer": layers}}
    fullmtypes = ["mtype1", "mtype2"]
    etypes = ["etype1"]
    _test_convert_emodel_etype_map(emodel_etype_map, fullmtypes, etypes,
                                   layers)


@pytest.mark.unit
def test_convert_emodel_etype_map_etype_regex():
    """prepare_combos.parse_files: test emodeletype map convert mtype regex"""
    layers = ["layer1", "2"]
    emodel_etype_map = {"emodel1": {"mm_recipe": "emodel2",
                                    "etype": ".*",
                                    "layer": layers}}
    fullmtypes = ["mtype1", "mtype2"]
    etypes = ["etype1"]
    _test_convert_emodel_etype_map(emodel_etype_map, fullmtypes, etypes,
                                   layers)


@pytest.mark.unit
def test_convert_emodel_etype_map_mtype_regex():
    """prepare_combos.parse_files: test emodel etype map convert etype regex"""

    layers = ["layer1", "2"]
    emodel_etype_map = {"emodel1": {"mm_recipe": "emodel2",
                                    "mtype": ".*",
                                    "layer": layers}}
    fullmtypes = ["mtype1", "mtype2"]
    etypes = ["etype1"]
    _test_convert_emodel_etype_map(emodel_etype_map, fullmtypes, etypes,
                                   layers)


@pytest.mark.unit
def test_convert_emodel_etype_map_morph_name_regex():
    """prepare_combos.parse_files: test emodel etype map convert morph regex"""

    layers = ["layer1", "2"]
    emodel_etype_map = {"emodel1": {"mm_recipe": "emodel2",
                                    "morph_name": ".*",
                                    "layer": layers}}
    fullmtypes = ["mtype1", "mtype2"]
    etypes = ["etype1"]
    _test_convert_emodel_etype_map(emodel_etype_map, fullmtypes, etypes,
                                   layers)
